# Basic Information
__title__     = "kocher_tools"
__version__   = '0.1.10'
__summary__   = "Bioinformatic Tools for the Kocher Lab"
__url__       = "https://kocher-guides.readthedocs.io/en/latest/"
__license__   = "MIT"
__copyright__ = "Copyright 2020, Andrew Webb and Sarah Kocher"

# Author Information
__author__    = "Andrew Webb and Sarah Kocher"
__email__     = "aewebb@princeton.edu"
